"""Module with core implementation of Server Object"""

from __future__ import annotations
from typing import Type, TypeVar, Literal
import asyncio
import functools
import logging
import threading

import websockets
from cbor2 import loads, dumps
import json

from .noodle_objects import *


# To allow for more flexible type annotation especially in create component
T = TypeVar("T", bound=Delegate)


def default_json_encoder(value):
    return str(value)


class Server(object):
    """NOODLES Server
    
    Attributes;
        clients (set): client connections
        custom_delegates (dict): 
            maps noodle object to its delegate class, empty by default
        delegates (dict):
            maps component ID to its delegate instance
        ids (dict): 
            maps object type to slot tracking info (next_slot, on_deck)
        components (dict):
            document's current state, contains all components with component ID as the key
        references (dict):
            maps component ID to all the component ID's that reference it
        delete_queue (set):
            components that are referenced but have been requested to be deleted
        message_map (dict):
            maps action and type to message ID
    """

    def __init__(self, port: int, starting_state: list[StartingComponent],
                 delegate_map: dict[Type[Delegate], Type[Delegate]] = None, json_output: str = None):
        """Constructor
        
        Args:
            port (int):
                port to run server on
            starting_state (list[StartingComponent]):
                list of objects containing the info to create components on initialization
            delegate_map (dict):
                maps noodles component type to instance of delegate class
            json_output (str):
                path to json file to output message logs
        Raises:
            Exception: Invalid Arguments or Method Not Specified when filling in starting state
        """

        self.port = port
        self.clients = set()
        self.ids = {}
        self.state = {}
        self.client_state = {}
        self.references = {}
        self.delete_queue = set()
        self.ready = threading.Event()
        self.shutdown_event = asyncio.Event()
        self.thread = None
        self.json_output = json_output
        if json_output:
            with open(json_output, "w") as outfile:  # Clear out old contents
                outfile.write("JSON Log\n")

        # Set up id's and custom delegates
        self.custom_delegates = delegate_map if delegate_map else {}
        self.id_map = id_map.copy()
        for old, new in self.custom_delegates.items():
            self.id_map[new] = self.id_map.pop(old)
        self.id_decoder = {val: key for key, val in id_map.items()}

        self.message_map = {
            ("create", MethodID): 0,
            ("delete", MethodID): 1,
            ("create", SignalID): 2,
            ("delete", SignalID): 3,
            ("create", EntityID): 4,
            ("update", EntityID): 5,
            ("delete", EntityID): 6,
            ("create", PlotID): 7,
            ("update", PlotID): 8,
            ("delete", PlotID): 9,
            ("create", BufferID): 10,
            ("delete", BufferID): 11,
            ("create", BufferViewID): 12,
            ("delete", BufferViewID): 13,
            ("create", MaterialID): 14,
            ("update", MaterialID): 15,
            ("delete", MaterialID): 16,
            ("create", ImageID): 17,
            ("delete", ImageID): 18,
            ("create", TextureID): 19,
            ("delete", TextureID): 20,
            ("create", SamplerID): 21,
            ("delete", SamplerID): 22,
            ("create", LightID): 23,
            ("update", LightID): 24,
            ("delete", LightID): 25,
            ("create", GeometryID): 26,
            ("delete", GeometryID): 27,
            ("create", TableID): 28,
            ("update", TableID): 29,
            ("delete", TableID): 30,
            ("update", None): 31,
            ("reset", None): 32,
            ("invoke", None): 33,
            ("reply", None): 34,
            ("initialized", None): 35
        }

        # Set up starting state
        for starting_component in starting_state:
            comp_type = starting_component.type
            comp_method = starting_component.method

            try:
                comp = self.create_component(comp_type, **starting_component.component_attrs)
            except Exception as e:
                raise TypeError(f"Invalid arguments to create {comp_type}: {e}")

            if comp_type == Method and comp_method:
                injected = InjectedMethod(self, comp_method)
                setattr(self, comp.name, injected)
            elif comp_type == Method and not comp_method:
                raise ValueError("Method not specified for starting method")

        logging.debug(f"Server initialized with objects: {self.state}")

    def __enter__(self):
        """Enter context manager"""
        self.thread = threading.Thread(target=self.run)
        self.thread.start()
        self.ready.wait()

        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """Exit context manager"""
        self.shutdown()
        self.thread.join()

    def run(self):
        """Run the server"""
        return asyncio.run(self.start_server())

    async def start_server(self):
        """Run the server and listen for connections"""
        # Create partial to pass server to handler
        handler = functools.partial(self.handle_client)

        logging.info("Starting up Server...")
        async with websockets.serve(handler, "", self.port):
            self.ready.set()
            while not self.shutdown_event.is_set():
                await asyncio.sleep(.1)

    def shutdown(self):
        """Shutdown the server"""
        logging.info("Shutting down server...")
        self.shutdown_event.set()

    def log_json(self, message: list):
        json_message = json.dumps(message, default=default_json_encoder)
        formatted_message = f"{json_message}\n"
        with open(self.json_output, "a") as outfile:
            outfile.write(formatted_message)

    async def send(self, websocket, message: list):
        """Send CBOR message using websocket

        Args:
            websocket (WebSocketClientProtocol):
                recipient of this data
            message (list):
                message to be sent, in list format
                [id, content, id, content...]
        """

        # Log message in json file if applicable
        if self.json_output:
            self.log_json(message)

        # Log message and send
        logging.debug(f"Sending Message: ID's {message[::2]}")
        await websocket.send(dumps(message))

    async def handle_client(self, websocket):
        """Coroutine for handling a client's connection

        Receives and delegates message handling to server

        args:
            websocket (WebSocketClientProtocol):
                client connection being handled
            server (Server):
                object for maintaining state of the scene
        """

        # Update track of clients
        self.clients.add(websocket)

        # Handle intro and update client
        raw_intro_msg = await websocket.recv()
        intro_msg = loads(raw_intro_msg)
        client_name = intro_msg[1]["client_name"]
        logging.info(f"Client '{client_name}' Connecting...")

        init_message = self.handle_intro()
        await self.send(websocket, init_message)

        # Listen for method invocation and keep all clients informed
        async for message in websocket:
            # Decode and log raw message
            message = loads(message)
            logging.debug(f"Message from client: {message}")

            # Handle the method invocation
            reply = self.handle_invoke(message[1])

            # Send method reply to client
            await self.send(websocket, list(reply))

        # Remove client if disconnected
        logging.debug(f"Client 'client_name' disconnected")
        self.clients.remove(websocket)

    def get_ids_by_type(self, component: Type[Delegate]) -> list:
        """Helper to get all ids for certain component type
        
        Args:
            component (type): type of component to get ID's for
        """

        return [key for key, val in self.state.items() if isinstance(val, component)]

    def get_delegate_id(self, name: str):
        """Helper to get a component with a type and name"""

        for delegate in self.state.values():
            if delegate.name == name:
                return delegate.id
        raise ValueError("No Component Found")

    def get_delegate(self, identifier: Union[ID, str, Dict[str, ID]]):
        """Getter for users to access components in state

        Can be called with an ID, name, or context of the delegate
        """
        if isinstance(identifier, ID):
            return self.state[identifier]
        elif isinstance(identifier, str):
            return self.state[self.get_delegate_id(identifier)]
        elif isinstance(identifier, dict):
            return self.get_delegate_by_context(identifier)
        else:
            raise TypeError(f"Invalid type for identifier: {type(identifier)}")

    def get_delegate_by_context(self, context: dict):
        """Helper to get a component by context"""

        entity = context.get("entity")
        table = context.get("table")
        plot = context.get("plot")
        if entity:
            return self.get_delegate(EntityID(*entity))
        elif table:
            return self.get_delegate(TableID(*table))
        elif plot:
            return self.get_delegate(PlotID(*plot))
        else:
            raise ValueError(f"Invalid context: {context}")

    def get_message_contents(self, action: str, noodle_object: NoodleObject, delta: set[str] = None):
        """Helper to handle construction of message dict
        
        Args:
            action (str): action taken with message
            noodle_object (NoodleObject): Delegate, Reply, or Invoke object
            delta (set): field names to be included in update
        """

        contents = {}
        if action == "create":
            base_delegate = self.id_decoder[type(noodle_object.id)]
            include = {field for field in base_delegate.__fields__ if field not in ["server", "signals"]}
            contents = noodle_object.dict(exclude_none=True, include=include)

        elif action == "invoke":
            contents = noodle_object.dict(exclude_none=True)

        elif action == "reply":

            if noodle_object.method_exception:
                e = noodle_object.method_exception
                contents = noodle_object.dict(exclude_none=True)
                contents["method_exception"] = {"code": e.code, "message": e.message, "data": e.data}
            else:
                contents = noodle_object.dict(exclude_none=True)

        elif action == "update":
            if not noodle_object:  # Document case
                contents["methods_list"] = self.get_ids_by_type(Method)
                contents["signals_list"] = self.get_ids_by_type(Signal)
            else:  # Normal update, include id, and any field in delta
                delta = set() if not delta else delta
                delta.add("id")
                contents = noodle_object.dict(exclude_none=True, include=delta)

        elif action == "delete":
            try:
                contents["id"] = noodle_object.id
            except AttributeError:
                raise Exception(f"Cannot delete a {noodle_object}")

        return contents

    def prepare_message(self, action: str, noodle_object: NoodleObject = None, delta: set[str] = None):
        """Given object and action, get id and message contents as dict

        Args:
            action (str): action taken with message
            noodle_object (NoodleObject): Component, Reply, or Invoke object
            delta (set): field names to be included in update
        """

        delegate_type = None if not isinstance(noodle_object, Delegate) else type(noodle_object.id)
        key = (action, delegate_type)
        message_id = self.message_map[key]
        contents = self.get_message_contents(action, noodle_object, delta)

        return message_id, contents

    def broadcast(self, message: list):
        """Broadcast message to all connected clients
        
        Args:
            message [tuple]: fully constructed message in form (tag/id, contents)
        """

        # Log message in json file if applicable
        if self.json_output:
            self.log_json(message)

        logging.debug(f"Broadcasting Message: ID's {message[::2]}")
        encoded = dumps(message)
        websockets.broadcast(self.clients, encoded)

    def handle_intro(self):
        """Formulate response for new client"""

        # Add create message for every object in state
        message = []
        ordered_components = order_components(self.state, self.references)
        for component in ordered_components:
            msg_id, content = self.prepare_message("create", component)
            message.extend([msg_id, content])

        # Add document update
        message.extend(self.prepare_message("update", None))

        # Finish with initialization message
        message.extend(self.prepare_message("initialized"))
        return message

    def handle_invoke(self, message: dict):
        """Handle all invokes coming from the client
        
        Take message and formulate response for clients. Tries to invoke and
        raises appropriate error codes if unsuccessful. Note that the method 
        technically doesn't raise any exceptions, instead the exception is 
        captured in a message and sent to the client.
        
        Args:
            message (dict): dict form of message from client
        """

        # Create generic reply with invalid invoke ID and attempt invoke
        reply_obj = Reply(invoke_id="-1")
        try:
            self.invoke_method(message, reply_obj)

        except Exception as e:
            if type(e) is MethodException:
                reply_obj.method_exception = e
            else:
                logging.error(f"\033[91mServerside Error: {e}\033[0m")
                reply_obj.method_exception = MethodException(code=-32603, message="Internal Error")

        return self.prepare_message("reply", reply_obj)

    def invoke_method(self, message: dict, reply: Reply):
        """Invoke method and build out reply object
        
        Mostly a helper for handle_invoke to raise proper method exceptions

        Args:
            message (dict): Invoke message in dict form
            reply (Reply): Practically empty reply object to be updated 
        """

        # Parse message
        try:
            method_id = MethodID(slot=message["method"][0], gen=message["method"][1])
            context = message.get("context")
            invoke_id = message["invoke_id"]
            args: list = message["args"]
            reply.invoke_id = invoke_id
        except Exception:
            raise MethodException(code=-32700, message="Parse Error")

        # Locate method
        try:
            method_name = self.state[method_id].name
            method = getattr(self, method_name)
        except Exception:
            raise MethodException(code=-32601, message="Method Not Found")

        # Invoke
        reply.result = method(context, *args)

    def update_references(self, parent_delegate: Delegate, current: NoodleObject, removing=False):
        """Update in-degree for all objects referenced by this one

        Recursively updates references for all components under a parent one. Here,
        the current object changes through the recursion while comp keeps track of 
        the parent. Essentially finds all the objects that delegate points to
        
        Args:
            parent_delegate (Delegate): parent component with new references to be tracked
            current (NoodleObject): current object being examined
            removing (bool): flag so function can be used to both add and remove references
        """

        for key, val in current:

            # Found a reference
            if key != "id" and isinstance(val, ID):
                if removing:
                    self.references[val].remove(parent_delegate.id)
                else:
                    self.references.setdefault(val, set()).add(parent_delegate.id)

            # Found another object to recurse on
            elif isinstance(val, NoodleObject):
                self.update_references(parent_delegate, val, removing)

            # Found list of objects or id's to recurse on 
            elif val is not None and isinstance(val, list):

                # Objects
                if len(val) > 0 and isinstance(val[0], NoodleObject):
                    for obj in val:
                        self.update_references(parent_delegate, obj, removing)

                # ID's
                elif len(val) > 0 and isinstance(val[0], ID):
                    for id in val:
                        if removing:
                            self.references[id].remove(parent_delegate.id)
                        else:
                            self.references.setdefault(id, set()).add(parent_delegate.id)

    def get_id(self, delegate_type: Type[Delegate]) -> ID:
        """Get next open ID
        
        Check for open slots then take the closest available slot

        Args:
            delegate_type (Type): type for desired ID
        """

        # Check if type is already tracked and set it up if not
        if delegate_type in self.ids:
            slot_info = self.ids[delegate_type]
        else:
            slot_info = SlotTracker()
            self.ids[delegate_type] = slot_info

        # Create the new ID from tracked info
        if slot_info.on_deck.empty():
            id_type = self.id_map[delegate_type]
            id = id_type(slot=slot_info.next_slot, gen=0)
            slot_info.next_slot += 1
            return id
        else:
            return slot_info.on_deck.get()

    # Interface methods to build server methods ===============================================

    def create_component(self, comp_type: Type[T], **kwargs) -> T:
        """Officially create new component in state
        
        This method updates state, updates references, and broadcasts msg to clients.
        It also handles the acquisition of a valid ID. This is a general creator method, but
        more specific versions exist for each component type
        
        Args:
            comp_type (Component Type): type of component to be created
            **kwargs: the user should specify the attributes of the component using 
                keyword arguments. Refer to the noodle objects to see which attributes
                are required and optional. Any deviation from the spec will raise a 
                validation exception. Note that since this method handles the ID, it 
                should not be specified as one of the keyword arguments.
        """

        # Get ID and try to create delegate from args
        comp_type = self.custom_delegates.get(comp_type, comp_type)
        comp_id = self.get_id(comp_type)
        try:
            new_delegate = comp_type(server=self, id=comp_id, **kwargs)
        except Exception as e:
            raise Exception(f"Args: {kwargs}, invalid for initializing a {comp_type}: {e}")

        # Update state and keep track of initial version for changes / update messages
        self.state[comp_id] = new_delegate
        self.client_state[comp_id] = new_delegate.copy()

        # Update references for each component referenced by this one
        self.update_references(new_delegate, new_delegate)

        # Create message and broadcast
        message = self.prepare_message("create", new_delegate)
        self.broadcast(message)

        # Return component or delegate instance if applicable
        return new_delegate

    def delete_component(self, delegate: Union[Delegate, ID]):
        """Delete object in state and update clients
        
        This method excepts a delegate, component, or component ID, and will attempt
        to delete the component as long as it is not referenced by any other component.
        If this component is still being used by another, it will be added to a queue so that
        it can be deleted later once that reference is no longer being used.

        Args:
            delegate (Component, Delegate, or ID): component / delegate to be deleted
        """

        # Handle cases so can except different input types - cast to ID
        if isinstance(delegate, Delegate):
            delegate = delegate
            del_id = delegate.id
        elif isinstance(delegate, ID):
            delegate = self.state[delegate]
            del_id = delegate.id
        else:
            raise TypeError(f"Invalid type for delegate when deleting: {type(delegate)}")

        # Delete if no references, or else queue it up for later
        if not self.references.get(del_id):
            self.broadcast(self.prepare_message("delete", delegate))
            del self.state[del_id]
            del self.client_state[del_id]

            # Free up the ID
            self.ids[type(delegate)].on_deck.put(del_id)

            # Clean out references from this object
            for refs in self.references.values():
                while del_id in refs:
                    refs.remove(del_id)

            # Check if anything in the queue is now clear to be deleted
            for comp_id in list(self.delete_queue):
                if not self.references.get(comp_id):
                    self.delete_queue.remove(comp_id)
                    self.delete_component(comp_id)

        else:
            logging.warning(f"Couldn't delete {delegate}, referenced by {self.references[del_id]}, added to queue")
            self.delete_queue.add(del_id)

    def find_delta(self, state, edited):
        """Helper to find differences between two objects
        
        Also checks to find recursive cases and cases where references
        should be updated, Does this get recursive deltas tho? should it?
        """

        delta = set()
        for field_name, value in edited:
            state_val = getattr(state, field_name)
            if value != state_val:
                delta.add(field_name)
        return delta

    def update_component(self, current: Delegate):
        """Update clients with changes to a component
        
        This method broadcasts changes to all clients including only fields
        specified in the set delta

        Args:
            current (Delegate): component that has been updated,
                should be a component with an update message
        """

        # Find difference between two states
        outdated = self.client_state[current.id]
        delta = self.find_delta(outdated, current)

        # Update references
        self.update_references(outdated, outdated, removing=True)
        self.update_references(current, current)

        # Update tracking state
        self.client_state[current.id] = current.copy()

        # Form message and broadcast
        try:
            message = self.prepare_message("update", current, delta)
            self.broadcast(message)
        except Exception as e:
            raise Exception(f"This obj can not be updated: {e}")

    def invoke_signal(self, signal: SignalID, on_component: Delegate, signal_data: list):
        """Send signal to target component
        
        Args:
            signal (ID): signal to be invoked
            on_component (Delegate): component to receive the signal
            signal_data (dict): 
        Takes Signal ID, on_component, and the data
        """

        # Get context from on_component
        if isinstance(on_component, Entity):
            context = InvokeIDType(entity=on_component.id)
        elif isinstance(on_component, Table):
            context = InvokeIDType(table=on_component.id)
        elif isinstance(on_component, Plot):
            context = InvokeIDType(plot=on_component.id)
        else:
            raise Exception(f"Invalid on_component type: {type(on_component)}")

        # Create invoke object and broadcast message
        invoke = Invoke(id=signal, context=context, signal_data=signal_data)
        message = self.prepare_message("invoke", invoke)
        self.broadcast(message)

    def create_method(self, name: str,
                      arg_doc: list[MethodArg],
                      doc: Optional[str] = None,
                      return_doc: Optional[str] = None) -> Method:
        return self.create_component(Method, name=name, doc=doc, return_doc=return_doc, arg_doc=arg_doc)

    def create_signal(self, name: str,
                      doc: Optional[str] = None,
                      arg_doc: list[MethodArg] = None) -> Signal:
        return self.create_component(Signal, name=name, doc=doc, arg_doc=arg_doc)

    def create_entity(self, name: Optional[str],
                      parent: Optional[EntityID] = None,
                      transform: Optional[Mat4] = None,
                      text_rep: Optional[TextRepresentation] = None,
                      web_rep: Optional[WebRepresentation] = None,
                      render_rep: Optional[RenderRepresentation] = None,
                      lights: Optional[list[LightID]] = None,
                      tables: Optional[list[TableID]] = None,
                      plots: Optional[list[PlotID]] = None,
                      tags: Optional[list[str]] = None,
                      methods_list: Optional[list[MethodID]] = None,
                      signals_list: Optional[list[SignalID]] = None,
                      influence: Optional[BoundingBox] = None) -> Entity:

        return self.create_component(Entity, name=name, parent=parent, transform=transform, text_rep=text_rep,
                                     web_rep=web_rep, render_rep=render_rep, lights=lights, tables=tables, plots=plots,
                                     tags=tags, methods_list=methods_list, signals_list=signals_list,
                                     influence=influence)

    def create_plot(self, name: Optional[str] = None,
                    table: Optional[TableID] = None,
                    simple_plot: Optional[str] = None,
                    url_plot: Optional[str] = None,
                    methods_list: Optional[list[MethodID]] = None,
                    signals_list: Optional[list[SignalID]] = None) -> Plot:

        return self.create_component(Plot, name=name, table=table, simple_plot=simple_plot, url_plot=url_plot,
                                     methods_list=methods_list, signals_list=signals_list)

    def create_buffer(self, name: Optional[str] = None,
                      size: int = None,
                      inline_bytes: bytes = None,
                      uri_bytes: str = None) -> Buffer:
        return self.create_component(Buffer, name=name, size=size, inline_bytes=inline_bytes, uri_bytes=uri_bytes)

    def create_bufferview(self,
                          source_buffer: BufferID,
                          offset: int,
                          length: int,
                          name: Optional[str] = None,
                          type: Literal["UNK", "GEOMETRY", "IMAGE"] = "UNK") -> BufferView:
        return self.create_component(BufferView, name=name, source_buffer=source_buffer,
                                     offset=offset, length=length, type=type)

    def create_material(self, name: Optional[str] = None,
                        pbr_info: Optional[PBRInfo] = PBRInfo(),
                        normal_texture: Optional[TextureRef] = None,
                        occlusion_texture: Optional[TextureRef] = None,
                        occlusion_texture_factor: Optional[float] = 1.0,
                        emissive_texture: Optional[TextureRef] = None,
                        emissive_factor: Optional[Vec3] = (1.0, 1.0, 1.0),
                        use_alpha: Optional[bool] = False,
                        alpha_cutoff: Optional[float] = .5,
                        double_sided: Optional[bool] = False) -> Material:
        return self.create_component(Material, name=name, pbr_info=pbr_info, normal_texture=normal_texture,
                                     occlusion_texture=occlusion_texture,
                                     occlusion_texture_factor=occlusion_texture_factor,
                                     emissive_texture=emissive_texture, emissive_factor=emissive_factor,
                                     use_alpha=use_alpha, alpha_cutoff=alpha_cutoff, double_sided=double_sided)

    def create_image(self, name: Optional[str] = None,
                     buffer_source: BufferID = None,
                     uri_source: str = None) -> Image:
        return self.create_component(Image, name=name, buffer_source=buffer_source, uri_source=uri_source)

    def create_texture(self, image: ImageID,
                       name: Optional[str] = None,
                       sampler: Optional[SamplerID] = None) -> Texture:
        return self.create_component(Texture, name=name, image=image, sampler=sampler)

    def create_sampler(self, name: Optional[str] = None,
                       mag_filter: Optional[Literal["NEAREST", "LINEAR"]] = "LINEAR",
                       min_filter: Optional[
                           Literal["NEAREST", "LINEAR", "LINEAR_MIPMAP_LINEAR"]] = "LINEAR_MIPMAP_LINEAR",
                       wrap_s: Optional[SamplerMode] = "REPEAT",
                       wrap_t: Optional[SamplerMode] = "REPEAT") -> Sampler:
        return self.create_component(Sampler, name=name, mag_filter=mag_filter, min_filter=min_filter,
                                     wrap_s=wrap_s, wrap_t=wrap_t)

    def create_light(self, name: Optional[str] = None,
                     color: Optional[RGB] = (1.0, 1.0, 1.0),
                     intensity: Optional[float] = 1.0,
                     point: PointLight = None,
                     spot: SpotLight = None,
                     directional: DirectionalLight = None) -> Light:
        return self.create_component(Light, name=name, color=color, intensity=intensity,
                                     point=point, spot=spot, directional=directional)

    def create_geometry(self, patches: list[GeometryPatch], name: Optional[str] = None) -> Geometry:
        return self.create_component(Geometry, name=name, patches=patches)

    def create_table(self, name: Optional[str] = None,
                     meta: Optional[str] = None,
                     methods_list: Optional[list[MethodID]] = None,
                     signals_list: Optional[list[SignalID]] = None) -> Table:
        return self.create_component(Table, name=name, meta=meta, methods_list=methods_list, signals_list=signals_list)


# Helpers for ordering messages
def top_sort_recurse(id, refs, visited, components, stack):
    """Helper for order_components to recurse"""

    visited[id] = True
    if id in refs:
        for ref in refs[id]:
            if not visited[ref]:
                top_sort_recurse(ref, refs, visited, components, stack)

    stack.append(components[id])


def order_components(components: dict[ID, Delegate],
                     refs: dict[ID, list[ID]]):
    """Helper for creating topological sort of components"""

    visited = {key: False for key in components}
    stack = []

    for id in components:
        if not visited[id]:
            top_sort_recurse(id, refs, visited, components, stack)

    return stack[::-1]



"""Python Server Library for NOODLES Protocol

This server library implements the NOODLES messaging protocol and provides objects for 
maintaining a scene in state. The server uses a websocket connection to send CBOR encoded 
messages. To customize its implementation, the library provides convenient interface 
methods to assist the user in writing their own methods for the server. The user can also 
add custom delegates to add additionally functionality to any of the standard components.

Modules:
    core.py
    geometry/
        geometry_creation.py
        geometry_objects.py
    delegates.py
    noodle_objects.py
    server.py
"""


__version__ = "0.2.1"

from .core import Server
from .noodle_objects import *

# Ensure that dependencies are installed for optional module
try:
    import numpy
    import meshio
    optionals = True
except ImportError:
    optionals = False

if optionals:
    from . import geometry



import socket
import re
import threading


class ByteServer(object):
    """Server to Host URI Bytes
    
    Attributes:
        host (str): IP address for server
        port (int): port server is listening on
        socket (socket): socket connection 
        buffers (dict): mapping tag to buffer
        next_tag (int): next available tag for a buffer
        url (str): base url to reach server without tag
        thread (Thread): background thread server is running in
        running (bool): flag indicating whether server is running
    """

    def __init__(self, port: int = 8000):
        """Constructor
        
        Args:
            port (int): port to listen on
        """

        name = socket.gethostname()
        try:  # supposed to work without .local, but had to add to match system preferences - sharing
            self.host = socket.gethostbyname(name)
        except socket.gaierror:
            self.host = socket.gethostbyname(f"{name}.local")

        self.port = port
        self.socket = None
        self.buffers = {}
        self.next_tag = 0
        self.url = f"http://{self.host}:{port}"

        self.thread = threading.Thread(target=self.run, args=())
        self.running = True
        self.thread.start()

    def get_tag(self):
        """Helper to get next tag"""

        tag = self.next_tag
        self.next_tag += 1
        return str(tag)

    def get_buffer(self, uri: str):
        """Helper to get bytes for a URI
        
        Mostly used in geometry creation for exporting as of right now

        Args:
            uri (str): uri for bytes
        """

        m = re.search(f'(?<={self.port}/).+\Z', uri)
        if m:
            tag = m.group(0)
            buffer_bytes = self.buffers[tag]
            return buffer_bytes
        else:
            raise Exception("Invalid HTTP Request")

    def run(self):
        """Main loop to run in thread
        
        Runs the server and listens for byte requests
        Uses HTTP protocol 
        """

        # Create socket
        server_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        server_socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        server_socket.bind((self.host, self.port))

        print(f"IP Address: {server_socket.getsockname()}")

        self.socket = server_socket

        server_socket.listen(1)
        print(f'Listening on port {self.port} ...')

        while self.running:
            # Wait for client connections
            print("Waiting for connection...")
            client_connection, client_address = server_socket.accept()

            # Get the client request
            request = client_connection.recv(1024).decode()
            print(f"Request: {request}")

            # Try to get tag from request with regex
            m = re.search('(?<=GET /)(.+?)(?= HTTP)', request)
            if m:
                tag = m.group(0)
                select_bytes = self.buffers[tag]
                header = f'HTTP/1.0 200 OK\r\nContent-Type: application/octet-stream\r\nContent-Length: {len(select_bytes)}\n\n'
                response = bytearray(header.encode()) + select_bytes
            else:
                header = f'HTTP/1.0 500 FAIL'
                response = header.encode()

            # Send HTTP response
            # client_connection.sendall(response.encode())
            # print(f'Response:\n{response}')
            client_connection.sendall(response)
            client_connection.close()

        # Clean up and close socket
        self.socket.close()

    def add_buffer(self, buffer) -> str:
        """Add buffer to server and return url to reach it
        
        Args:
            buffer (bytes): bytes to add as buffer
        """

        tag = self.get_tag()
        self.buffers[tag] = buffer
        url = f"{self.url}/{tag}"
        print(f"Adding Buffer: {url}")

        return url

    def shutdown(self):
        """Stop running thread"""

        self.running = False
        self.thread.join()


def main():
    """Main method for testing"""

    server = ByteServer(port=60000)
    server.add_buffer('HTTP/1.0 200 OK\n\nTEST TEST TEST'.encode())
    server.add_buffer(b'TESTING BYTES')


if __name__ == "__main__":
    main()



"""Geometry Creation Library for Noodles Protocol

This library provides assistance with the creation of complex linked geometry components.
Users can use these built in methods as opposed to constructing buffers, buffer views,
entities, and geometries manually.
"""

from .geometry_creation import build_geometry_patch, build_entity, create_instances, update_entity, add_instances, \
    geometry_from_mesh, export_mesh
from .geometry_objects import AttributeInput, GeometryPatchInput
from .byte_server import ByteServer



"""Module for assisting with the creation of geometry objects"""
from math import sqrt
from collections import deque
from statistics import mean
from typing import Optional, Tuple
import numpy as np
import meshio

from .. import noodle_objects as nooobs
from ..core import Server

from .geometry_objects import AttributeInput, GeometryPatchInput
from .byte_server import ByteServer

INLINE_LIMIT = 10000

SIZES = {
    # in bytes
    "U8": 1,
    "U16": 2,
    "U32": 4,
    "U8VEC4": 4,
    "U16VEC2": 4,
    "VEC2": 8,
    "VEC3": 12,
    "VEC4": 16,
    "MAT3": 36,
    "MAT4": 64
}

FORMAT_MAP = {
    "U8": np.int8,
    "U16": np.int16,
    "U32": np.int32,
    "U8VEC4": np.int8,
    "U16VEC2": np.int16,
    "VEC2": np.single,
    "VEC3": np.single,
    "VEC4": np.single,
    "MAT3": np.single,
    "MAT4": np.single
}

DEFAULT_POSITION = [0.0, 0.0, 0.0, 1.0]
DEFAULT_COLOR = [1.0, 1.0, 1.0, 1.0]
DEFAULT_ROTATION = [0.0, 0.0, 0.0, 1.0]
DEFAULT_SCALE = [1.0, 1.0, 1.0, 1.0]


def get_format(num_vertices: int) -> str:
    """Helper to get format that can accommodate number of vertices
    
    Args:
        num_vertices (int): number of vertices needed to store in mesh
    """

    if num_vertices < 256:
        return 'U8'
    elif num_vertices < 65536:
        return 'U16'
    else:
        return 'U32'


def set_up_attributes(patch_input: GeometryPatchInput, generate_normals: bool):
    """Constructs attribute info from input type
    
    Takes list input and constructs objects 
    that can be used in build_geometry_patch

    Args:
        patch_input (GeometryPatchInput): stores lists of vertices, indices
            index type, material, and possibly normals, tangents,
            textures, and colors
        generate_normals (bool): calculate normals for mesh or not
    """

    # Generate normals if not indicated in input
    if not patch_input.normals and generate_normals:
        patch_input.normals = calculate_normals(patch_input.vertices, patch_input.indices)

    # Add attribute info based on the input lists
    attribute_info = []
    position = AttributeInput(
        semantic="POSITION",
        format="VEC3",
        normalized=False,
    )
    attribute_info.append(position)

    normal = AttributeInput(
        semantic="NORMAL",
        format="VEC3",
        normalized=False,
    )
    attribute_info.append(normal)

    if patch_input.tangents:
        tangent = AttributeInput(
            semantic="TANGENT",
            format="VEC3",
            normalized=False,
        )
        attribute_info.append(tangent)

    if patch_input.textures:
        texture = AttributeInput(
            semantic="TEXTURE",
            format="U16VEC2",
            normalized=True,
        )
        attribute_info.append(texture)

    if patch_input.colors:
        color = AttributeInput(
            semantic="COLOR",
            format="U8VEC4",
            normalized=True,
        )
        attribute_info.append(color)

        # Check color format and correct
        if any(i > 1 for i in patch_input.colors[0]):
            for i in range(len(patch_input.colors)):
                patch_input.colors[i] = [x / 255 for x in patch_input.colors[i]]

    # Use input to get offsets
    offset = 0
    for attribute in attribute_info:
        attribute.offset = offset
        offset += SIZES[attribute.format]

    # Use final offset to get stride
    for attribute in attribute_info:
        attribute.stride = offset

    return attribute_info


def build_geometry_buffer(server: Server, name, patch_input: GeometryPatchInput, index_format: str,
                          attribute_info: list[AttributeInput],
                          byte_server: ByteServer = None) -> Tuple[nooobs.Buffer, int]:
    """Builds a buffer component

    Args:
        server (Server): server to create component on
        name (str): name to give component
        patch_input (GeometryPatchInput): lists of attributes and point data
        index_format (str): format the indices should take
        attribute_info (list[AttributeInput]): Info on the attributes, mostly used for formatting
        byte_server (ByteServer): byte server to use if needed
    """

    # Filter out inputs unspecified by user, and group attributes by point
    data = [x for x in
            [patch_input.vertices, patch_input.normals, patch_input.tangents, patch_input.textures, patch_input.colors]
            if x]
    points = zip(*data)

    # Build byte array by iterating through points and their attributes
    buffer_bytes = bytearray(0)
    for point in points:
        for info, attr in zip(point, attribute_info):
            attr_size = FORMAT_MAP[attr.format]
            new_bytes = np.array(info, dtype=attr_size).tobytes(order='C')
            buffer_bytes.extend(new_bytes)

    # Add index bytes to byte array
    index_offset = len(buffer_bytes)
    index_bytes = np.array(patch_input.indices, dtype=FORMAT_MAP[index_format]).tobytes(order='C')
    buffer_bytes.extend(index_bytes)

    # Create buffer component using uri bytes if needed
    size = len(buffer_bytes)
    if size > INLINE_LIMIT:
        print(f"Large Mesh: Using URI Bytes")
        uri = byte_server.add_buffer(buffer_bytes)
        buffer = server.create_buffer(name=name, size=size, uri_bytes=uri)
        return buffer, index_offset
    else:
        buffer = server.create_buffer(
            name=name,
            size=size,
            inline_bytes=buffer_bytes
        )
        return buffer, index_offset


def build_geometry_patch(server: Server, name: str, patch_input: GeometryPatchInput,
                         byte_server: ByteServer = None, generate_normals: bool = True) -> nooobs.GeometryPatch:
    """Build a Geometry Patch with related buffers and views
    
    Args:
        server (Server): server to create components on
        name (str): name for the components
        patch_input (GeometryPatch): input lists with data to create the patch
        byte_server (ByteServer): optional server to use if mesh is larger than 10Kb
        generate_normals (bool): whether to calculate normals for this mesh
    """

    # Set up some constants
    vert_count = len(patch_input.vertices)
    index_count = len(patch_input.indices) * len(patch_input.indices[0])
    index_format = get_format(vert_count)

    # Set up attributes with given lists
    attribute_info = set_up_attributes(patch_input, generate_normals=generate_normals)

    # Build buffer with given lists
    buffer: nooobs.Buffer
    buffer, index_offset = build_geometry_buffer(server, name, patch_input, index_format, attribute_info, byte_server)

    # Make buffer view component
    buffer_view: nooobs.BufferView = server.create_component(
        nooobs.BufferView,
        name=name,
        source_buffer=buffer.id,
        type="GEOMETRY",
        offset=0,  # What is this? cant always assume 0
        length=buffer.size
    )

    # Create attribute objects from buffer view and attribute info
    attributes = []
    for attribute in attribute_info:
        attr_obj = nooobs.Attribute(view=buffer_view.id, **dict(attribute))
        attributes.append(attr_obj)

    # Make index to describe indices at end of buffer
    index = nooobs.Index(
        view=buffer_view.id,
        count=index_count,
        offset=index_offset,
        format=index_format
    )

    # Finally create patch 
    patch = nooobs.GeometryPatch(
        attributes=attributes,
        vertex_count=vert_count,
        indices=index,
        type=patch_input.index_type,
        material=patch_input.material
    )

    return patch


def build_instance_buffer(server: Server, name: str, matrices: list[nooobs.Mat4]) -> nooobs.Buffer:
    """Build Buffer from Mat4 to Represent Instances
    
    Args:
        server (Server): server to put buffer component on
        name (str): name for the buffer
        matrices (Mat4): instance matrices
    """

    buffer_bytes = np.array(matrices, dtype=np.single).tobytes()

    buffer = server.create_component(
        nooobs.Buffer,
        name=f"Instance buffer for {name}",
        size=len(buffer_bytes),
        inline_bytes=buffer_bytes
    )

    return buffer


def build_entity(server: Server, geometry: nooobs.Geometry, instances: list[list] = None):
    """Build Entity from Geometry
    
    Args:
        server (Server): server to build entity component on
        geometry (Geometry): geometry to link entity to
        instances (Mat4): optional instance matrix, can use create_instances to generate
    """

    # Set name to match geometry
    name = geometry.name if geometry.name else None

    # Create instance buffer and view if specified
    if instances:
        buffer = build_instance_buffer(server, name, instances)
        buffer_view = server.create_component(
            nooobs.BufferView,
            name=f"Instance View for {name}",
            source_buffer=buffer.id,
            type="UNK",
            offset=0,
            length=buffer.size
        )
        instance = nooobs.InstanceSource(view=buffer_view.id, stride=0, bb=None)
    else:
        instance = None

    # Create render rep and entity from geometry and instances
    rep = nooobs.RenderRepresentation(mesh=geometry.id, instances=instance)
    entity = server.create_component(nooobs.Entity, name=name, render_rep=rep)

    return entity


def create_instances(
        positions: list[list] = None,
        colors: list[list] = None,
        rotations: list[list] = None,
        scales: list[list] = None) -> list[list]:
    """Create new instances for an entity
    
    All lists are optional and will be filled with defaults
    By default one instance is created at least
    Lists are padded out to 4 values

    Args:
        positions (list[Vec3]): positions for each instance
        colors (list[Vec4]): Colors for each instance
        rotations (list[Vec4]): Rotations for each instance
        scales (list[Vec3]): Scales for each instance
    """

    def padded(lst: list, default_val: float = 1.0):
        """Helper to pad the lists"""

        lst = list(lst)
        if len(lst) < 4:
            lst += [default_val] * (4 - len(lst))
        return lst

    # Safeguard against None values for input
    if rotations is None:
        rotations = []
    if colors is None:
        colors = []
    if scales is None:
        scales = []

    # If no inputs specified create one default instance
    if not positions:
        positions = [DEFAULT_POSITION]

    # Use the longest input as number of instances
    num_instances = max([len(l) for l in [positions, colors, rotations, scales]])

    # Build the matrices and extend
    instances = []
    for i in range(num_instances):
        position = padded(positions[i]) if i < len(positions) else DEFAULT_POSITION
        color = padded(colors[i]) if i < len(colors) else DEFAULT_COLOR
        rotation = padded(rotations[i]) if i < len(rotations) else DEFAULT_ROTATION
        scale = padded(scales[i]) if i < len(scales) else DEFAULT_SCALE
        instances.extend([position, color, rotation, scale])

    return instances


def update_entity(server: Server, entity: nooobs.Entity, geometry: nooobs.Geometry = None, instances: list = None):
    """Update an entity with new instances or geometry
    
    Args:
        server (Server): server with entity to update
        entity (Entity): Entity to be updated
        geometry (Geometry): Optional new geometry if that is being changed
        instances (list[Mat4]): Optional new instances if that is changed
    """

    # Get name from entity if applicable
    name = entity.name if entity.name else None

    # Get render rep and ensure entity is working with geometry
    old_rep = entity.render_rep
    if not old_rep:
        raise Exception("Entity isn't renderable")

    # Set geometry id based on whether there is new geometry or not
    if geometry:
        mesh = geometry.id
    else:
        mesh = old_rep.mesh

    # Build new buffer / view for instances or use existing instances
    if instances:
        buffer = build_instance_buffer(server, name, instances)
        buffer_view = server.create_component(
            nooobs.BufferView,
            name=f"Instance View for {name}",
            source_buffer=buffer.id,
            type="UNK",
            offset=0,
            length=buffer.size
        )
        instance = nooobs.InstanceSource(view=buffer_view.id, stride=0, bb=None)
    else:
        instance = old_rep.instances

    # Create new render rep for entity and update entity
    rep = nooobs.RenderRepresentation(mesh=mesh, instances=instance)
    entity.render_rep = rep
    entity = server.update_component(entity)

    # Clean up old components with deletes
    if instances:
        server.delete_component(server.state[old_rep.instances.view].source_buffer)
        server.delete_component(old_rep.instances.view)
    elif geometry:
        server.delete_component(old_rep.mesh)

    return entity


def add_instances(server: Server, entity: nooobs.Entity, instances: list):
    """Add instances to an existing entity
    
    Args:
        server (Server): server with entity to update
        entity (Entity): entity to be updated
        instances (list[Mat4]): new instances to be added, can be generated using 
            create_instances()
    """

    # Ensure we're working with an entity that can be rendered
    try:
        rep = entity.render_rep

        # Get old instance buffer from entity's render rep
        old_view: nooobs.BufferView = server.state[rep.instances.view]
        old_buffer: nooobs.Buffer = server.state[old_view.source_buffer]
        old_instances = np.frombuffer(old_buffer.inline_bytes, dtype=np.single)

        # Combine new and old instances
        combined = np.append(old_instances, instances)
        update_entity(server, entity, instances=combined.tolist())
    except:
        raise Exception("Entity isn't renderable")


# ---------------------------------- Mesh Importing ----------------------------------#

def convert(color: int):
    """Helper to convert decimal pymesh values to RGBA array"""

    rgb = []
    for i in range(4):
        rgb.append(color % 256)
        color = color // 256

    alpha = rgb.pop(0)
    rgb.append(alpha)
    return rgb


def meshlab_load(server: Server, byte_server: ByteServer, file,
                 material: nooobs.Material, mesh_name: Optional[str] = None, generate_normals: bool = True):
    """Use pymeshlab to load types unsupported by meshio

    This method is only called from geometry_from_mesh if needed
    
    Args:   
        server (Server): server to load geometry onto
        byte_server (ByteServer): server to support URI bytes if needed
        file (str, path): file to load mesh from
        material (Material): material to use in geometry
        mesh_name (str): optional name
        generate_normals (bool): whether to calculate normals for the mesh
    """
    import pymeshlab

    ms = pymeshlab.MeshSet()
    ms.load_new_mesh(file)
    mesh = ms.current_mesh()
    # mesh.compute_normal_per_vertex("Simple Average")
    print(f"Finished Loading Mesh...")

    # Extract data from mesh set structure
    vertices = mesh.vertex_matrix().tolist()
    indices = mesh.face_matrix().tolist()
    # normals = mesh.vertex_normal_matrix().tolist() # Look like they could be off
    normals = None
    tangents = None  # TBD
    textures = None  # TBD
    colors = [convert(color) for color in mesh.vertex_color_array().tolist()]

    # Create patch / geometry for point geometry
    patches = []
    patch_info = GeometryPatchInput(
        vertices=vertices,
        indices=indices,
        index_type="TRIANGLES",
        material=material.id,
        normals=normals,
        tangents=tangents,
        textures=textures,
        colors=colors)
    patches.append(build_geometry_patch(server, mesh_name, patch_info, byte_server, generate_normals=generate_normals))
    geometry = server.create_component(nooobs.Geometry, name=mesh_name, patches=patches)

    return geometry


def geometry_from_mesh(server: Server, file, material: nooobs.Material,
                       mesh_name: Optional[str] = None, byte_server: ByteServer = None, generate_normals: bool = True):
    """Construct geometry from mesh file
    
    Can specify byte server if it is a big mesh and needs uri bytes
    """

    # Create meshio mesh object to extract data from file, if unsupported file type use meshlab
    try:
        mesh = meshio.read(file)
    except Exception as e:
        return meshlab_load(server, byte_server, file, material, mesh_name, generate_normals=generate_normals)

        # Define Meshio helper functions

    def get_point_attr(mesh_obj, attr: str):
        """Helper to get attribute from mesh object"""

        data = mesh_obj.point_data.get(attr)
        try:
            return data.tolist()
        except:
            return data

    def get_triangles(mesh_obj):
        """Helper to get triangles from mesh object"""

        for cell in mesh_obj.cells:
            if cell.type == "triangle":
                return cell.data.tolist()
        return None

    vertices = mesh.points.tolist()
    indices = get_triangles(mesh)

    # I think attribute name is going to be specific to mesh - how to map to right attribute?
    normals = get_point_attr(mesh, "NORMAL")
    tangents = get_point_attr(mesh, "TANGENT")
    textures = get_point_attr(mesh, "TEXTURE")
    colors = get_point_attr(mesh, "COLOR")

    # Create patch / geometry for point geometry
    patches = []
    patch_info = GeometryPatchInput(
        vertices=vertices,
        indices=indices,
        index_type="TRIANGLES",
        material=material.id,
        normals=normals,
        tangents=tangents,
        textures=textures,
        colors=colors)
    patches.append(build_geometry_patch(server, mesh_name, patch_info, byte_server, generate_normals=generate_normals))
    geometry = server.create_component(nooobs.Geometry, name=mesh_name, patches=patches)

    return geometry


def export_mesh(server: Server, geometry: nooobs.Geometry, new_file_name: str, byte_server: ByteServer = None):
    """Export noodles geometry to mesh file
    
    Args:
        server (Server): server to get components from
        geometry (Geometry): geometry to export
        new_file_name (str): name for file being created
        byte_server (ByteServer): server storing bytes if applicable 
    """

    points = []
    indices = []
    point_data = {}
    for patch in geometry.patches:

        # Extract info from patch
        index = patch.indices
        view: nooobs.BufferView = server.get_delegate(index.view)
        buffer = server.get_delegate(view.source_buffer)
        inline, uri = buffer.inline_bytes, buffer.uri_bytes
        if inline:
            geo_bytes = inline
        else:
            try:
                geo_bytes = byte_server.get_buffer(uri)
            except:
                raise Exception("No byte server specified for uri byte mesh")

        # Reconstruct indices from buffer
        raw_indices = np.frombuffer(geo_bytes, dtype=FORMAT_MAP[index.format], count=index.count, offset=index.offset)
        grouped = [list(x) for x in zip(*(iter(raw_indices),) * 3)]
        indices.extend(grouped)

        # Reconstruct points and attributes from buffer
        i = 0
        attribute_bytes = geo_bytes[:index.offset]
        while i < len(attribute_bytes):
            for attribute in patch.attributes:
                current_chunk = attribute_bytes[i:i + SIZES[attribute.format]]
                attr_name = attribute.semantic
                attr_data = np.frombuffer(current_chunk, dtype=FORMAT_MAP[attribute.format]).tolist()
                i += SIZES[attribute.format]
                if attr_name == "POSITION":
                    points.append(attr_data)
                else:
                    point_data.setdefault(attr_name, []).append(attr_data)

                    # Construct mesh and export
    mesh = meshio.Mesh(
        points,
        [("triangle", indices)],
        point_data=point_data)
    mesh.write(new_file_name)


def dot_product(v1, v2):
    """Helper to take dot product"""
    product = 0
    for c1, c2 in zip(v1, v2):
        product += (round(c1, 5) * round(c2, 5))
    return product


def calculate_normals(vertices: list[list], indices: list[list]):
    """TODO"""

    # Idea: go through all the triangles, and calculate normal for each one and attach average to each vertex
    print(f"Generating normals for {len(vertices)} vertices")
    normals = {}
    adjacents = {}
    for triangle in indices:

        # Calculate the normal
        v1, v2, v3 = triangle
        p1, p2, p3 = vertices[v1], vertices[v2], vertices[v3]
        vector1 = np.array([p2[0] - p1[0], p2[1] - p1[1], p2[2] - p1[2]])  # p1 -> p2
        vector2 = np.array([p3[0] - p1[0], p3[1] - p1[1], p3[2] - p1[2]])  # p1 -> p3
        normal = np.cross(vector1, vector2)

        # Attach normal to each vertex
        for vertex in triangle:

            existing_normals = normals.get(vertex)

            # Add in new normals with matching orientation
            if existing_normals:
                if dot_product(existing_normals[0], normal) < 0:
                    existing_normals.append([-x for x in normal])
                    print(f"Mismatching triangle normals @ {vertex}")
                else:
                    existing_normals.append(normal)
            else:
                normals[vertex] = [normal]

            # Mark other vertices in triangle as adjacent
            other_vert = [x for x in triangle if x != vertex]
            adjacents.setdefault(vertex, set()).update(other_vert)

    print(f"Vertices {len(vertices)} vs. Normals {len(normals)}")

    # Find averages
    for vertex, normal_list in normals.items():

        # Calculate average
        average_normal = []
        for component in zip(*normal_list):
            average_normal.append(mean(component))

        # Normalize and set normal to keep track of final average 
        length = sqrt((average_normal[0] ** 2) + (average_normal[1] ** 2) + (average_normal[2] ** 2))
        normals[vertex] = [x / length for x in average_normal]

    # Orient normals to match
    print(f"Vertices {len(vertices)} vs. Normals {len(normals)}")

    center = [mean(x) for x in zip(*vertices)]
    visited = set()
    starting_index = indices[0][0]
    discovered = deque()  # (index, neighbor)
    discovered.append((starting_index, starting_index))
    while discovered:

        # print(f"Orienting normal {len(visited)}/{len(discovered)}")
        current_index, neighbor = discovered.popleft()
        dot = dot_product(normals[current_index], normals[neighbor])
        if dot < 0:  # Flip vector
            normals[current_index] = [-x for x in normals[current_index]]

        for adjacent in adjacents[current_index]:
            if adjacent not in visited and (adjacent, current_index) not in discovered:
                discovered.append((adjacent, current_index))

        visited.add(current_index)

    print(f"Vertices {len(vertices)} vs. Normals {len(normals)}")
    # Find number pointing towards center
    num_inward = 0
    for index, normal in normals.items():
        center_vector = [x - y for x, y in zip(vertices[index], center)]
        if np.dot(normal, center_vector) < 0:
            num_inward += 1

    # If majority are inward invert
    if num_inward > (len(vertices) / 2):
        for key, normal in normals.items():
            normals[key] = [-x for x in normal]

    print(f"Finished getting normals...\nNum Inward: {num_inward}")
    print(f"Vertices {len(vertices)} vs. Normals {len(normals)}")
    return [normals.get(i, [0, 0, 0]) for i in range(len(vertices))]



"""Collection of Noodles Objects

Follows the specification in the cddl document, and
implements strict validation
"""

from enum import Enum
from math import pi
from queue import Queue
from typing import Callable, Optional, Any, Union, List, Tuple, Dict, NamedTuple

from pydantic import BaseModel, root_validator

""" =============================== ID's ============================= """


class ID(NamedTuple):
    slot: int
    gen: int

    def compact_str(self):
        return f"|{self.slot}/{self.gen}|"

    def __str__(self):
        return f"{type(self).__name__}{self.compact_str()}"

    def __key(self):
        return type(self), self.slot, self.gen

    def __eq__(self, other: object) -> bool:
        if type(other) is type(self):
            return self.__key() == other.__key()
        else:
            return False

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash(self.__key())


class MethodID(ID):
    pass


class SignalID(ID):
    pass


class EntityID(ID):
    pass


class PlotID(ID):
    pass


class BufferID(ID):
    pass


class BufferViewID(ID):
    pass


class MaterialID(ID):
    pass


class ImageID(ID):
    pass


class TextureID(ID):
    pass


class SamplerID(ID):
    pass


class LightID(ID):
    pass


class GeometryID(ID):
    pass


class TableID(ID):
    pass


""" ====================== Generic Parent Class ====================== """


class NoodleObject(BaseModel):
    """Parent Class for all noodle objects"""

    class Config:
        """Configuration for Validation"""

        arbitrary_types_allowed = True
        use_enum_values = True


class Delegate(NoodleObject):
    """Parent class for all delegates

    Defines general methods that should be available for all delegates

    Attributes:
        server (Server): server delegate is attached to
        id: (ID): Unique identifier for delegate
        name (str): Name of delegate
        signals (dict): Signals that can be called on delegate, method name to callable
    """

    server: object  # Better way to annotate this without introducing circular imports?
    id: ID = None
    name: Optional[str] = "No-Name"
    signals: Optional[dict] = {}

    def __str__(self):
        return f"{self.name} - {type(self).__name__} - {self.id.compact_str()}"


""" ====================== Common Definitions ====================== """

Vec3 = Tuple[float, float, float]
Vec4 = Tuple[float, float, float, float]
Mat3 = Tuple[float, float, float,
             float, float, float,
             float, float, float]
Mat4 = Tuple[float, float, float, float,
             float, float, float, float,
             float, float, float, float,
             float, float, float, float]

RGB = Vec3
RGBA = Vec4


class AttributeSemantic(Enum):
    position = "POSITION"
    normal = "NORMAL"
    tangent = "TANGENT"
    texture = "TEXTURE"
    color = "COLOR"


class Format(Enum):
    u8 = "U8"
    u16 = "U16"
    u32 = "U32"
    u8vec4 = "U8VEC4"
    u16vec2 = "U16VEC2"
    vec2 = "VEC2"
    vec3 = "VEC3"
    vec4 = "VEC4"
    mat3 = "MAT3"
    mat4 = "MAT4"


class IndexFormat(str, Enum):
    u8 = "U8"
    u16 = "U16"
    u32 = "U32"


class PrimitiveType(Enum):
    points = "POINTS"
    lines = "LINES"
    line_loop = "LINE_LOOP"
    line_strip = "LINE_STRIP"
    triangles = "TRIANGLES"
    triangle_strip = "TRIANGLE_STRIP"


class ColumnType(str, Enum):
    text = "TEXT"
    real = "REAL"
    integer = "INTEGER"


class BufferType(str, Enum):
    unknown = "UNK"
    geometry = "GEOMETRY"
    image = "IMAGE"


class SamplerMode(Enum):
    clamp_to_edge = "CLAMP_TO_EDGE"
    mirrored_repeat = "MIRRORED_REPEAT"
    repeat = "REPEAT"


class MagFilterTypes(Enum):
    nearest = "NEAREST"
    linear = "LINEAR"


class MinFilterTypes(Enum):
    nearest = "NEAREST"
    linear = "LINEAR"
    linear_mipmap_linear = "LINEAR_MIPMAP_LINEAR"


class SelectionRange(NoodleObject):
    key_from_inclusive: int
    key_to_exclusive: int


class Selection(NoodleObject):
    name: str
    rows: Optional[List[int]] = None
    row_ranges: Optional[List[SelectionRange]] = None


class MethodArg(NoodleObject):
    name: str
    doc: Optional[str] = None
    editor_hint: Optional[str] = None


class BoundingBox(NoodleObject):
    min: Vec3
    max: Vec3


class TextRepresentation(NoodleObject):
    txt: str
    font: Optional[str] = "Arial"
    height: Optional[float] = .25
    width: Optional[float] = -1.0


class WebRepresentation(NoodleObject):
    source: str
    height: Optional[float] = .5
    width: Optional[float] = .5


class InstanceSource(NoodleObject):
    view: BufferViewID  # view of mat4
    stride: int
    bb: Optional[BoundingBox] = None


class RenderRepresentation(NoodleObject):
    mesh: GeometryID
    instances: Optional[InstanceSource] = None


class TextureRef(NoodleObject):
    texture: TextureID
    transform: Optional[Mat3] = [1.0, 0.0, 0.0,
                                 0.0, 1.0, 0.0,
                                 0.0, 0.0, 1.0, ]
    texture_coord_slot: Optional[int] = 0.0


class PBRInfo(NoodleObject):
    base_color: RGBA = [1.0, 1.0, 1.0, 1.0]
    base_color_texture: Optional[TextureRef] = None  # assume SRGB, no premult alpha

    metallic: Optional[float] = 1.0
    roughness: Optional[float] = 1.0
    metal_rough_texture: Optional[TextureRef] = None  # assume linear, ONLY RG used


class PointLight(NoodleObject):
    range: float = -1.0


class SpotLight(NoodleObject):
    range: float = -1.0
    inner_cone_angle_rad: float = 0.0
    outer_cone_angle_rad: float = pi / 4


class DirectionalLight(NoodleObject):
    range: float = -1.0


class Attribute(NoodleObject):
    view: BufferViewID
    semantic: AttributeSemantic
    channel: Optional[int] = None
    offset: Optional[int] = 0
    stride: Optional[int] = 0
    format: Format
    minimum_value: Optional[List[float]] = None
    maximum_value: Optional[List[float]] = None
    normalized: Optional[bool] = False


class Index(NoodleObject):
    view: BufferViewID
    count: int
    offset: Optional[int] = 0
    stride: Optional[int] = 0
    format: IndexFormat


class GeometryPatch(NoodleObject):
    attributes: List[Attribute]
    vertex_count: int
    indices: Optional[Index] = None
    type: PrimitiveType
    material: MaterialID  # Material ID


class InvokeIDType(NoodleObject):
    entity: Optional[EntityID] = None
    table: Optional[TableID] = None
    plot: Optional[PlotID] = None

    @root_validator
    def one_of_three(cls, values):
        already_found = False
        for field in values:
            if values[field] and already_found:
                raise ValueError("More than one field entered")
            elif values[field]:
                already_found = True

        if not already_found:
            raise ValueError("No field provided")
        else:
            return values


class TableColumnInfo(NoodleObject):
    name: str
    type: ColumnType


class TableInitData(NoodleObject):
    columns: List[TableColumnInfo]
    keys: List[int]
    data: List[List[Any]]  # Originally tried union, but currently order is used to coerce by pydantic
    selections: Optional[List[Selection]] = None

    # too much overhead? - strict mode
    @root_validator
    def types_match(cls, values):
        for row in values['data']:
            for col, i in zip(values['columns'], range(len(row))):
                text_mismatch = isinstance(row[i], str) and col.type != "TEXT"
                real_mismatch = isinstance(row[i], float) and col.type != "REAL"
                int_mismatch = isinstance(row[i], int) and col.type != "INTEGER"
                if text_mismatch or real_mismatch or int_mismatch:
                    raise ValueError(f"Column Info doesn't match type in data: {col, row[i]}")
        return values


""" ====================== NOODLE COMPONENTS / DELEGATES ====================== """


class Method(Delegate):
    id: MethodID
    name: str
    doc: Optional[str] = None
    return_doc: Optional[str] = None
    arg_doc: List[MethodArg] = []


class Signal(Delegate):
    id: SignalID
    name: str
    doc: Optional[str] = None
    arg_doc: List[MethodArg] = None


class Entity(Delegate):
    id: EntityID
    name: Optional[str] = None

    parent: Optional[EntityID] = None
    transform: Optional[Mat4] = None

    text_rep: Optional[TextRepresentation] = None
    web_rep: Optional[WebRepresentation] = None
    render_rep: Optional[RenderRepresentation] = None

    lights: Optional[List[LightID]] = None
    tables: Optional[List[TableID]] = None
    plots: Optional[List[PlotID]] = None
    tags: Optional[List[str]] = None
    methods_list: Optional[List[MethodID]] = None
    signals_list: Optional[List[SignalID]] = None

    influence: Optional[BoundingBox] = None


class Plot(Delegate):
    id: PlotID
    name: Optional[str] = None

    table: Optional[TableID] = None

    simple_plot: Optional[str] = None
    url_plot: Optional[str] = None

    methods_list: Optional[List[MethodID]] = None
    signals_list: Optional[List[SignalID]] = None

    @root_validator
    def one_of(cls, values):
        if bool(values['simple_plot']) != bool(values['url_plot']):
            return values
        else:
            raise ValueError("One plot type must be specified")


class Buffer(Delegate):
    id: BufferID
    name: Optional[str] = None
    size: int = None

    inline_bytes: bytes = None
    uri_bytes: str = None

    @root_validator
    def one_of(cls, values):
        if bool(values['inline_bytes']) != bool(values['uri_bytes']):
            return values
        else:
            raise ValueError("One plot type must be specified")


class BufferView(Delegate):
    id: BufferViewID
    name: Optional[str] = None
    source_buffer: BufferID

    type: BufferType = BufferType.unknown
    offset: int
    length: int


class Material(Delegate):
    id: MaterialID
    name: Optional[str] = None

    pbr_info: Optional[PBRInfo] = PBRInfo()
    normal_texture: Optional[TextureRef] = None

    occlusion_texture: Optional[TextureRef] = None  # assumed to be linear, ONLY R used
    occlusion_texture_factor: Optional[float] = 1.0

    emissive_texture: Optional[TextureRef] = None  # assumed to be SRGB, ignore A
    emissive_factor: Optional[Vec3] = [1.0, 1.0, 1.0]

    use_alpha: Optional[bool] = False
    alpha_cutoff: Optional[float] = .5

    double_sided: Optional[bool] = False


class Image(Delegate):
    id: ImageID
    name: Optional[str] = None

    buffer_source: BufferID = None
    uri_source: str = None

    @root_validator
    def one_of(cls, values):
        if bool(values['buffer_source']) != bool(values['uri_source']):
            return values
        else:
            raise ValueError("One plot type must be specified")


class Texture(Delegate):
    id: TextureID
    name: Optional[str] = None
    image: ImageID  # Image ID
    sampler: Optional[SamplerID] = None


class Sampler(Delegate):
    id: SamplerID
    name: Optional[str] = None

    mag_filter: Optional[MagFilterTypes] = MagFilterTypes.linear
    min_filter: Optional[MinFilterTypes] = MinFilterTypes.linear_mipmap_linear

    wrap_s: Optional[SamplerMode] = "REPEAT"
    wrap_t: Optional[SamplerMode] = "REPEAT"


class Light(Delegate):
    id: LightID
    name: Optional[str] = None

    color: Optional[RGB] = [1.0, 1.0, 1.0]
    intensity: Optional[float] = 1.0

    point: PointLight = None
    spot: SpotLight = None
    directional: DirectionalLight = None

    @root_validator
    def one_of(cls, values):
        already_found = False
        for field in ['point', 'spot', 'directional']:
            if values[field] and already_found:
                raise ValueError("More than one field entered")
            elif values[field]:
                already_found = True

        if not already_found:
            raise ValueError("No field provided")
        else:
            return values


class Geometry(Delegate):
    id: GeometryID
    name: Optional[str] = None
    patches: List[GeometryPatch]


class Table(Delegate):
    id: TableID
    name: Optional[str] = None

    meta: Optional[str] = None
    methods_list: Optional[List[MethodID]] = None
    signals_list: Optional[List[SignalID]] = None

    def handle_insert(self, new_rows: list[list[int]]):
        pass

    def handle_update(self, keys: list[int], rows: list[list[int]]):
        pass

    def handle_delete(self, keys: list[int]):
        pass

    def handle_clear(self):
        pass

    def handle_set_selection(self, selection: Selection):
        pass

    # Signals ---------------------------------------------------
    def table_reset(self, tbl_init: TableInitData):
        """Invoke table reset signal"""
        pass

    def table_updated(self, keys: list[int], rows: list[list[int]]):
        pass

    def table_rows_removed(self, keys: list[int]):
        pass

    def table_selection_updated(self, selection: Selection):
        pass


""" ====================== Communication Objects ====================== """


class Invoke(NoodleObject):
    id: SignalID
    context: Optional[InvokeIDType] = None  # if empty - document
    signal_data: List[Any]


# Note: this isn't technically an exception
# for now this uses a model so that it can be validated / sent as message easier
# Fix to inherit from Exception, maybe use separate model for sending
# class MethodException(NoodleObject):
#     code: int
#     message: Optional[str] = None
#     data: Optional[Any] = None
class MethodException(Exception):
    def __init__(self, code: int, message: Optional[str] = None, data: Optional[Any] = None):
        self.code = code
        self.message = message
        self.data = data


class Reply(NoodleObject):
    invoke_id: str
    result: Optional[Any] = None
    method_exception: Optional[MethodException] = None


""" ====================== Miscellaneous Objects ====================== """

id_map = {
    Method: MethodID,
    Signal: SignalID,
    Table: TableID,
    Plot: PlotID,
    Entity: EntityID,
    Material: MaterialID,
    Geometry: GeometryID,
    Light: LightID,
    Image: ImageID,
    Texture: TextureID,
    Sampler: SamplerID,
    Buffer: BufferID,
    BufferView: BufferViewID
}


class InjectedMethod(object):
    """Representation of user Injected Method
    
    Note that the call method automatically inserts a server
    reference as an argument to all user defined functions
    """

    def __init__(self, server, method) -> None:
        self.server = server
        self.method = method

    def __call__(self, *args, **kwargs):
        return self.method(self.server, *args, **kwargs)


class SlotTracker(object):
    """Object to keep track of next available slot
    
    Next slot is next unused slot while on_deck
    keeps track of slots that have opened up
    """

    def __init__(self):
        self.next_slot = 0
        self.on_deck = Queue()


class StartingComponent(object):
    """User input object for setting starting components on server"""

    def __init__(self, kind, component_attrs: Dict[str, Any], method: Optional[Callable] = None):
        self.type = kind
        self.component_attrs = component_attrs
        self.method = method


def get_context(delegate):
    """Helper to get context from delegate"""

    if isinstance(delegate, Entity):
        return {"entity": delegate.id}
    elif isinstance(delegate, Table):
        return {"table": delegate.id}
    elif isinstance(delegate, Plot):
        return {"plot": delegate.id}
    else:
        return None



"""Module with additional objects to help with geometry creation

These are based on the noodle_objects.py module and implement validation
"""

from typing import Optional

from .. import noodle_objects as nooobs

class AttributeInput(nooobs.NoodleObject):
    """Input for setting attributes of a buffer 
    
    User should not have to concern themselves with this input
    """

    semantic: nooobs.AttributeSemantic
    format: nooobs.Format
    normalized: bool
    offset: Optional[int]
    stride: Optional[int]


class GeometryPatchInput(nooobs.NoodleObject):
    """User input object for creating a geometry patch"""

    vertices: list
    indices: list
    index_type: str 
    material: nooobs.MaterialID
    normals: Optional[list] 
    tangents: Optional[list]
    textures: Optional[list] 
    colors: Optional[list]


